package Centurion.lib.jazzlib;

public final class Inflater
{
    private static final int[] CPLENS;
    private static final int[] CPLEXT;
    private static final int[] CPDIST;
    private static final int[] CPDEXT;
    private int mode;
    private int readAdler;
    private int neededBits;
    private int repLength;
    private int repDist;
    private int uncomprLen;
    private boolean isLastBlock;
    private long totalOut;
    private long totalIn;
    private boolean nowrap;
    private StreamManipulator input;
    private OutputWindow outputWindow;
    private InflaterDynHeader dynHeader;
    private InflaterHuffmanTree litlenTree;
    private InflaterHuffmanTree distTree;
    private Adler32 adler;
    
    public Inflater() {
        this(false);
    }
    
    public Inflater(final boolean nowrap) {
        super();
        this.nowrap = nowrap;
        this.adler = new Adler32();
        this.input = new StreamManipulator();
        this.outputWindow = new OutputWindow();
        this.mode = (nowrap ? 2 : 0);
    }
    
    public final boolean finished() {
        return this.mode == 12 && this.outputWindow.getAvailable() == 0;
    }
    
    public final int getRemaining() {
        return this.input.getAvailableBytes();
    }
    
    public final int getTotalIn() {
        return (int)(this.totalIn - this.getRemaining());
    }
    
    public final int getTotalOut() {
        return (int)this.totalOut;
    }
    
    public final int inflate(final byte[] buf, int off, int len) throws DataFormatException {
        if (len == 0) {
            return 0;
        }
        if (0 > off || off > off + len || off + len > buf.length) {
            throw new ArrayIndexOutOfBoundsException();
        }
        int count = 0;
        do {
            if (this.mode != 11) {
                final int more = this.outputWindow.copyOutput(buf, off, len);
                this.adler.update(buf, off, more);
                off += more;
                count += more;
                this.totalOut += more;
                if ((len -= more) == 0) {
                    return count;
                }
                continue;
            }
        } while (this.decode() || (this.outputWindow.getAvailable() > 0 && this.mode != 11));
        return count;
    }
    
    public final boolean needsDictionary() {
        return this.mode == 1 && this.neededBits == 0;
    }
    
    public final boolean needsInput() {
        return this.input.needsInput();
    }
    
    public final void reset() {
        this.mode = (this.nowrap ? 2 : 0);
        final long n = 0L;
        this.totalOut = n;
        this.totalIn = n;
        this.input.reset();
        this.outputWindow.reset();
        this.dynHeader = null;
        this.litlenTree = null;
        this.distTree = null;
        this.isLastBlock = false;
        this.adler.reset();
    }
    
    public final void setInput(final byte[] buf, final int off, final int len) {
        this.input.setInput(buf, off, len);
        this.totalIn += len;
    }
    
    private boolean decodeHeader() throws DataFormatException {
        final int header;
        if ((header = this.input.peekBits(16)) < 0) {
            return false;
        }
        this.input.dropBits(16);
        final int n;
        if ((n = ((header << 8 | header >> 8) & 0xFFFF)) % 31 != 0) {
            throw new DataFormatException("Header checksum illegal");
        }
        if ((n & 0xF00) != 0x800) {
            throw new DataFormatException("Compression Method unknown");
        }
        if ((n & 0x20) == 0x0) {
            this.mode = 2;
        }
        else {
            this.mode = 1;
            this.neededBits = 32;
        }
        return true;
    }
    
    private boolean decodeDict() {
        while (this.neededBits > 0) {
            final int dictByte;
            if ((dictByte = this.input.peekBits(8)) < 0) {
                return false;
            }
            this.input.dropBits(8);
            this.readAdler = (this.readAdler << 8 | dictByte);
            this.neededBits -= 8;
        }
        return false;
    }
    
    private boolean decodeHuffman() throws DataFormatException {
        int free = this.outputWindow.getFreeSpace();
        while (free >= 258) {
            Label_0158: {
                switch (this.mode) {
                    case 7: {
                        int symbol;
                        while (((symbol = this.litlenTree.getSymbol(this.input)) & 0xFFFFFF00) == 0x0) {
                            this.outputWindow.write(symbol);
                            if (--free < 258) {
                                return true;
                            }
                        }
                        if (symbol >= 257) {
                            try {
                                this.repLength = Inflater.CPLENS[symbol - 257];
                                this.neededBits = Inflater.CPLEXT[symbol - 257];
                            }
                            catch (ArrayIndexOutOfBoundsException ex) {
                                throw new DataFormatException("Illegal rep length code");
                            }
                            break Label_0158;
                        }
                        if (symbol < 0) {
                            return false;
                        }
                        this.distTree = null;
                        this.litlenTree = null;
                        this.mode = 2;
                        return true;
                    }
                    case 8:
                        if (this.neededBits > 0) {
                            this.mode = 8;
                            final int i;
                            if ((i = this.input.peekBits(this.neededBits)) < 0) {
                                return false;
                            }
                            this.input.dropBits(this.neededBits);
                            this.repLength += i;
                        }
                        this.mode = 9;
                    case 9: {
                        final int symbol;
                        if ((symbol = this.distTree.getSymbol(this.input)) < 0) {
                            return false;
                        }
                        try {
                            this.repDist = Inflater.CPDIST[symbol];
                            this.neededBits = Inflater.CPDEXT[symbol];
                        }
                        catch (ArrayIndexOutOfBoundsException ex2) {
                            throw new DataFormatException("Illegal rep dist code");
                        }
                    }
                    case 10:
                        if (this.neededBits > 0) {
                            this.mode = 10;
                            final int i;
                            if ((i = this.input.peekBits(this.neededBits)) < 0) {
                                return false;
                            }
                            this.input.dropBits(this.neededBits);
                            this.repDist += i;
                        }
                        this.outputWindow.repeat(this.repLength, this.repDist);
                        free -= this.repLength;
                        this.mode = 7;
                        continue;
                    default:
                        throw new IllegalStateException();
                }
            }
        }
        return true;
    }
    
    private boolean decodeChksum() throws DataFormatException {
        while (this.neededBits > 0) {
            final int chkByte;
            if ((chkByte = this.input.peekBits(8)) < 0) {
                return false;
            }
            this.input.dropBits(8);
            this.readAdler = (this.readAdler << 8 | chkByte);
            this.neededBits -= 8;
        }
        if ((int)this.adler.getValue() != this.readAdler) {
            throw new DataFormatException("Adler chksum doesn't match: " + Integer.toHexString((int)this.adler.getValue()) + " vs. " + Integer.toHexString(this.readAdler));
        }
        this.mode = 12;
        return false;
    }
    
    private boolean decode() throws DataFormatException {
        switch (this.mode) {
            case 0:
                return this.decodeHeader();
            case 1:
                return this.decodeDict();
            case 11:
                return this.decodeChksum();
            case 2:
                if (this.isLastBlock) {
                    if (this.nowrap) {
                        this.mode = 12;
                        return false;
                    }
                    this.input.skipToByteBoundary();
                    this.neededBits = 32;
                    this.mode = 11;
                    return true;
                }
                else {
                    final int type;
                    if ((type = this.input.peekBits(3)) < 0) {
                        return false;
                    }
                    this.input.dropBits(3);
                    if ((type & 0x1) != 0x0) {
                        this.isLastBlock = true;
                    }
                    switch (type >> 1) {
                        case 0:
                            this.input.skipToByteBoundary();
                            this.mode = 3;
                            break;
                        case 1:
                            this.litlenTree = InflaterHuffmanTree.defLitLenTree;
                            this.distTree = InflaterHuffmanTree.defDistTree;
                            this.mode = 7;
                            break;
                        case 2:
                            this.dynHeader = new InflaterDynHeader();
                            this.mode = 6;
                            break;
                        default:
                            throw new DataFormatException("Unknown block type " + type);
                    }
                    return true;
                }
            case 3: {
                final int nlen = this.input.peekBits(16);
                this.uncomprLen = nlen;
                if (nlen < 0) {
                    return false;
                }
                this.input.dropBits(16);
                this.mode = 4;
            }
            case 4: {
                final int nlen;
                if ((nlen = this.input.peekBits(16)) < 0) {
                    return false;
                }
                this.input.dropBits(16);
                if (nlen != (this.uncomprLen ^ 0xFFFF)) {
                    throw new DataFormatException("broken uncompressed block");
                }
                this.mode = 5;
            }
            case 5:
                this.uncomprLen -= this.outputWindow.copyStored(this.input, this.uncomprLen);
                if (this.uncomprLen == 0) {
                    this.mode = 2;
                    return true;
                }
                return !this.input.needsInput();
            case 6:
                if (!this.dynHeader.decode(this.input)) {
                    return false;
                }
                this.litlenTree = this.dynHeader.buildLitLenTree();
                this.distTree = this.dynHeader.buildDistTree();
                this.mode = 7;
                return this.decodeHuffman();
            case 7:
            case 8:
            case 9:
            case 10:
                return this.decodeHuffman();
            case 12:
                return false;
            default:
                throw new IllegalStateException();
        }
    }
    
    static {
        CPLENS = new int[] { 3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31, 35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258 };
        CPLEXT = new int[] { 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0 };
        CPDIST = new int[] { 1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193, 257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145, 8193, 12289, 16385, 24577 };
        CPDEXT = new int[] { 0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13 };
    }
}
